/* Caprice32 - Amstrad CPC Emulator
   (c) Copyright 1997-2004 Ulrich Doewich

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cap32.h"
#include "crtc.h"
#include "z80.h"

extern t_CPC CPC;
extern t_CRTC CRTC;
extern t_GateArray GateArray;
extern t_VDU VDU;
extern t_z80regs z80;

extern byte *pbRAM;
extern byte mode0_table[512], mode1_table[1024];

#ifdef DEBUG_CRTC
extern FILE *pfoDebug;
#endif





void draw16bpp_border_half_NULL(void)
{
   CPC.scr_offs += 8; // update PC screen buffer address
}

void draw16bpp_border_half(void)
{
   int colour;
   register word *mem_ptr;

   colour = GateArray.palette[16];
   mem_ptr = CPC.scr_base3 + CPC.scr_offs; // PC screen buffer address
   *(mem_ptr++) = colour; // write one pixel of border colour
   *(mem_ptr++) = colour;
   *(mem_ptr++) = colour;
   *(mem_ptr++) = colour;

   *(mem_ptr++) = colour;
   *(mem_ptr++) = colour;
   *(mem_ptr++) = colour;
   *(mem_ptr++) = colour;

   CPC.scr_offs += 8; // update PC screen buffer address
}

void draw16bpp_mode0_half(dword addr)
{
   byte idx;
   register word *mem_ptr;
   //dword val;
   register byte *col_ptr;

   mem_ptr = CPC.scr_base3 + CPC.scr_offs; // PC screen buffer address
   idx = *(pbRAM + addr); // grab first CPC screen memory byte
   col_ptr = &mode0_table[(idx<<1)+0];

	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];

	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];

   idx = *(pbRAM + ((addr+1)&0xffff)); // grab second CPC screen memory byte
   col_ptr = &mode0_table[(idx<<1)+0];

	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];

	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr)];

   CPC.scr_offs += 8; // update PC screen buffer address
}


void draw16bpp_mode1_half(dword addr)
{
   byte idx;
   register word *mem_ptr;
   //reg_pair val;
 
   mem_ptr = CPC.scr_base3 + CPC.scr_offs; // PC screen buffer address
   idx = *(pbRAM + addr); // grab first CPC screen memory byte

   register byte *col_ptr;
   col_ptr = &mode1_table[(idx<<2)+0];

	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*col_ptr];
	   
   idx = *(pbRAM + ((addr+1)&0xffff)); // grab second CPC screen memory byte
   col_ptr = &mode1_table[(idx<<2)+0];
   
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*(col_ptr++)];
	   *(mem_ptr++) = GateArray.palette[*col_ptr];

   CPC.scr_offs += 8; // update PC screen buffer address
}


void draw16bpp_mode2_half(dword addr)
{
   byte pat;
   register word *mem_ptr;
   dword pen_on, pen_off;
   reg_pair val;

   mem_ptr = CPC.scr_base3 + CPC.scr_offs; // PC screen buffer address
   pen_on = GateArray.palette[1];
   pen_off = GateArray.palette[0];

   pat = *(pbRAM + addr); // grab first CPC screen memory byte
   val.w.l = (pat & 0x80) ? pen_on : pen_off;
   val.w.h = (pat & 0x40) ? pen_on : pen_off;
   *(mem_ptr++) = val.d; // write four pixels

   val.w.l = (pat & 0x20) ? pen_on : pen_off;
   val.w.h = (pat & 0x10) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   val.w.l = (pat & 0x08) ? pen_on : pen_off;
   val.w.h = (pat & 0x04) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   val.w.l = (pat & 0x02) ? pen_on : pen_off;
   val.w.h = (pat & 0x01) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   pat = *(pbRAM + ((addr+1)&0xffff)); // grab second CPC screen memory byte
   val.w.l = (pat & 0x80) ? pen_on : pen_off;
   val.w.h = (pat & 0x40) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   val.w.l = (pat & 0x20) ? pen_on : pen_off;
   val.w.h = (pat & 0x10) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   val.w.l = (pat & 0x08) ? pen_on : pen_off;
   val.w.h = (pat & 0x04) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   val.w.l = (pat & 0x02) ? pen_on : pen_off;
   val.w.h = (pat & 0x01) ? pen_on : pen_off;
   *(mem_ptr++) = val.d;

   CPC.scr_offs += 8; // update PC screen buffer address
}
