/*! \file interface.c
 *  \brief Gui Driver.
 *         Dingoo Library.
 *
 *  \version 0.9
 *
 *  Dingoo Interface/Gui driver
 *
 *   Pituka - Dingoo Port
 *  (c) Copyright 2009 David Colmenero (aka D_Skywalk)
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/fb.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <linux/input.h>

#include "../caprice/cap32.h"
#include "../caprice/crtc.h"

#include "../global.h"
#include "images/images.h"

#include "dngutils.h"
extern unsigned short *screen_base;
extern int key_dev;
extern t_VDU VDU;
extern t_CPC CPC;
extern int emuDone;
extern int EscDone;
t_img splash;
t_img keyboard;

static Goo_cursor puntero = { 8, 2, 90, 207};
Goo_cursor cfg_tecla = { 0, 0, 270, 32};

unsigned short *pled_screen;

extern char spool_cad[20];
extern bool spool;

extern struct input_event KeyEv;

byte cpc_key=0, cpc_key_push=0xff;

//extern unsigned char botones_usr[11];
extern Bitu8 keyboard_matrix[16];
extern Bitu8 bit_values[8];

//------------------------------------------------------

int InitGFX (void)
{

  InitImg(&keyboard);

  if(!LoadBitmap(&keyboard, screenkeyb))
     return 0;

  pled_screen = screen_base + (WRES_GOO*5) + (WRES_GOO - 16);

  return 1;

}

void FreeGFX (void)
{
  free(&keyboard);
}

void ShowSplash (void)
{
  InitImg(&splash);

  if(!LoadBitmap(&splash, screensplash))
     return;

  DrawBitmap (&splash, 0);

  usleep(500); //wait vsync

  DrawBitmap (&splash, 0);

  sleep(3);

  FreeImg (&splash);

}

//------------------------------------------------------

static void ShowCursor (void);
//static void ShowKeyPad (void);
void input_bind(Bitu8 key, int pos);

unsigned char tabla_teclado[90]={
0x82,0x82,0x80,0x81,0x71,0x70,0x61,0x60,0x51,0x50,0x41,0x40,0x31,0x30,0x97,0x12,0x13,0x03,
0x84,0x84,0x83,0x73,0x72,0x62,0x63,0x53,0x52,0x43,0x42,0x33,0x32,0x21,0x22,0x24,0x14,0x04,
0x86,0x86,0x85,0x74,0x75,0x65,0x64,0x54,0x55,0x45,0x44,0x35,0x34,0x23,0x22,0x15,0x16,0x05,
0x25,0x25,0x87,0x77,0x76,0x67,0x66,0x56,0x46,0x47,0x37,0x36,0x26,0x25,0x25,0x17,0x00,0x07,
0x27,0x27,0x11,0x11,0x57,0x57,0x57,0x57,0x57,0x57,0x57,0x57,0x06,0x06,0x06,0x10,0x02,0x01
};

inline void Keyboard_PressKey(byte KeyCode, bool pressed)
{
  if (KeyCode == 0xff)
     return;

  if(pressed)
  {
     keyboard_matrix[KeyCode >> 4] &= ~bit_values[KeyCode & 7]; // tecla pulsada
     cpc_key_push=KeyCode;
  }
  else
     keyboard_matrix[KeyCode >> 4] |= bit_values[KeyCode & 7];  // tecla soltada

}

void Keyboard_getControl (void)
{

  if((!KeyEv.value))
  {
      switch (KeyEv.code) {
          case KEY_ESC:
              EscDone = 0;
              break;

          case KEY_TAB: Keyboard_PressKey(0x25, 0); break;
          case KEY_BACKSPACE: Keyboard_PressKey(0x97, 0); break;
          case KEY_LEFTCTRL: Keyboard_PressKey(cpc_key_push, 0); break;
     }
     return;
  }

  switch (KeyEv.code) {
     case KEY_LEFT:
         if(puntero.x > 0){
             puntero.px-=10;
             puntero.x-=1;
         }else{
             puntero.px=180;
             puntero.x=17;
         }
         break;
     case KEY_RIGHT:
         if(puntero.x < 17){
             puntero.px+=10;
             puntero.x+=1;
         }else{
             puntero.px=20;
             puntero.x=1;
         }
         break;
     case KEY_UP:
         if(puntero.y > 0){
             puntero.py-=10;
             puntero.y-=1;
         }else{
             puntero.py=227;
             puntero.y=4;
         }
         break;
     case KEY_DOWN:
         if(puntero.y < 4){
             puntero.py+=10;
             puntero.y+=1;
         }else{
             puntero.py=187;
             puntero.y=0;
         }
         break;

     case KEY_ESC: 
	      CPC.control = 0;
              VDU.vstart = 45;
              VDU.vheight = CPC_VISIBLE_SCR_HEIGHT;
              EscDone = 1;
	      break;

     case KEY_ENTER: 
              if(EscDone) emuDone = 1;
    #if 0
              if(cfg_tecla.x < 6)
                  cfg_tecla.x += 1;
              else
                  cfg_tecla.x = 0;
    #endif

              break;
     case KEY_TAB: Keyboard_PressKey(0x25, 1); break;
     case KEY_BACKSPACE: Keyboard_PressKey(0x97, 1); break;

     case KEY_LEFTCTRL:
              Keyboard_PressKey(tabla_teclado[((puntero.y*18)+puntero.x)], 1);
              break;
     #if 0 //no funciona estropea el array de button
     case KEY_LEFTALT:
              spool=1;
              //botones[cfg_tecla.x]=tabla_teclado[((puntero.y*18)+puntero.x)];
              input_bind(tabla_teclado[((puntero.y*18)+puntero.x)], cfg_tecla.x);
              sprintf(spool_cad,"You have remapped a key (%i)", (int) cfg_tecla.x);
     #endif
     default: return;
  }

}

inline void ShowKeyboard (void)
{
  
  DrawBitmap (&keyboard, 176);

  //mostrar cursores
  ShowCursor();
  //ShowKeyPad();

}

#define LEDCOLOR 0x7FE

inline void ShowLed (void)
{

  memset(pled_screen,LEDCOLOR, sizeof(unsigned short)*8);
  memset(pled_screen+(WRES_GOO*1),LEDCOLOR, sizeof(unsigned short)*8);
  memset(pled_screen+(WRES_GOO*2),LEDCOLOR, sizeof(unsigned short)*8);
  memset(pled_screen+(WRES_GOO*3),LEDCOLOR, sizeof(unsigned short)*8);

}

#define CURCOLOR 0xfcc0

static void ShowCursor (void)
{
  unsigned short *ptr_screen = screen_base + (puntero.py * WRES_GOO) + puntero.px;

  memset(ptr_screen,CURCOLOR, sizeof(unsigned short)*4);
  memset(ptr_screen+(WRES_GOO*1),CURCOLOR, sizeof(unsigned short)*3);
  memset(ptr_screen+(WRES_GOO*2),CURCOLOR, sizeof(unsigned short)*2);
  memset(ptr_screen+(WRES_GOO*3),CURCOLOR, sizeof(unsigned short)*1);

}
#if 0
static void ShowKeyPad (void)
{
  
  switch(cfg_tecla.x){
	
    case DINGOO_JOY_A:
        cfg_tecla.px=236;
        cfg_tecla.py=142;
        break;
    case DINGOO_JOY_B:
        cfg_tecla.px=230;
        cfg_tecla.py=136;
        break;
    case DINGOO_JOY_START:
        cfg_tecla.px=230;
        cfg_tecla.py=124;
        break;
    case DINGOO_JOY_UP:
        cfg_tecla.px=210;
        cfg_tecla.py=141;
        break;
    case DINGOO_JOY_RIGHT:
        cfg_tecla.px=211;
        cfg_tecla.py=140;
        break;
    case DINGOO_JOY_DOWN:
        cfg_tecla.px=210;
        cfg_tecla.py=136;
        break;
    case DINGOO_JOY_LEFT:
        cfg_tecla.px=206;
        cfg_tecla.py=140;
        break;
  }

  unsigned short *ptr_screen = screen_base + (cfg_tecla.py * WRES_GOO) + cfg_tecla.px;

  memset(ptr_screen,CURCOLOR, sizeof(unsigned short)*2);
  memset(ptr_screen+(WRES_GOO*1),CURCOLOR, sizeof(unsigned short)*2);

}
#endif


