/*! \file main.c
 *  \brief Main
 *         Dingoo Library.
 *
 *  \version 0.9
 *
 *  Dingoo main src
 *
 *   Pituka - Dingoo Port
 *  (c) Copyright 2009 David Colmenero (aka D_Skywalk)
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/fb.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <linux/input.h>
#include <unistd.h>

#include "../caprice/cap32.h"
#include "../caprice/crtc.h"

#include "../global.h"
#include "../port/gfx/pituka-arranque.h"
#include "dngutils.h"
#include "interface.h"

#define DEFAULT_FIFO_SIZE	(256*1024)

/*** 2D Video Globals ***/
void *pix = NULL;

extern Bitu8 pituka_arranque[];
extern Bitu16 pituka_pal[];

extern Bitu8 *pbSndBuffer;
extern Bitu8 *pbSndBufferEnd;
extern Bitu8 *pbSndStream;
extern t_CPC CPC;
extern t_VDU VDU;
extern t_FDC FDC;
extern t_GateArray GateArray;
extern int emuDone;

unsigned short pituka_surface[320*240];

SDL_Surface *screen;
int screen_bps;
unsigned short *screen_base;

int fb_dev;
int key_dev;
static int con_dev = -1;
static int con_old_vt = -1;

WiituKa_Status WiiStatus = {0,0,0,0,0,0,0,0,0,0};
xmlWiiCFG WiitukaXML;

/* FUNCIONES */
int init_buffer (void);
void close_buffer (void);


/* GLOBALES */
t_WiiRom globalRom;


Goo_gun gunstick;
char extension[5];
char file_state[13];
char rom_name[256];

Pituka_SpoolKeys spoolkeys;
unsigned char reiniciado;
bool pinta_menu;

char debugt [1024];
char current_path [1024] = "/CPCROMS/";

char spool_cad[20] = "";
bool spool = 0;
bool spool_act = 0;

bool sonido_ant;

/* CPC */
extern Bitu8 keyboard_matrix[16];
extern Bitu8 bit_values[8];
extern Bitu8 keyboard_translation[320];
extern Bitu8 keyboard_translation_SDL[320];

extern Bitu8 * pbGPBuffer;

unsigned char * my_splash;

struct input_event KeyEv;

int EscDone = 0;

Bitu8 button[11]={
0x90, //UP
0x93,//RIGHT
0x91,//DW
0x92,//LEFT
0x95,//B
0x94,//A
0x22,//START
0x85,//X
0x57,//Y
0x80,//RL
0x81//RR
};


static int init_dingoo (void)
{

  char vt_name[128];

  if(SDL_Init(SDL_INIT_TIMER) < 0) //
     return 0;

  //desactivar consola
  snprintf(vt_name, sizeof(vt_name), "%s%d", "/dev/tty", 1);
  int base_con_dev = open(vt_name, O_WRONLY);

  if (base_con_dev != -1) {
      int con_num;

      if (ioctl(base_con_dev, VT_OPENQRY, &con_num) != -1) {
          snprintf(vt_name, sizeof(vt_name), "%s%d", "/dev/tty", con_num);
          con_dev = open(vt_name, O_WRONLY);

          if (con_dev != -1) {
              struct vt_stat vts;
              if (ioctl(con_dev, VT_GETSTATE, &vts) == 0) con_old_vt = vts.v_active;

              if (ioctl(con_dev, VT_ACTIVATE, con_num) != -1
                  && ioctl(con_dev, VT_WAITACTIVE, con_num) != -1
                  && ioctl(con_dev, KDSETMODE, KD_GRAPHICS) != -1) goto con_success;
            }
      }

      if (con_dev != -1) {
          close(con_dev);
          con_dev = -1;
      }

      static const char cmd[] = "\033[?1;0;0c";
      write(base_con_dev, cmd, sizeof(cmd)-1);

  con_success:
      close(base_con_dev);
  }

  //iniciar controles
  if ((key_dev = open("/dev/event0", O_RDONLY | O_NONBLOCK)) == -1) return -1;

  SDL_JoystickOpen(0);
  SDL_ShowCursor(SDL_DISABLE);

  //iniciar video (framebuffer)
  fb_dev = open("/dev/fb0", O_RDWR);
  if(fb_dev)
  {
    screen_base = mmap(NULL, sizeof(unsigned short)*WRES_GOO*HRES_GOO, PROT_READ | PROT_WRITE, MAP_SHARED, fb_dev, 0);
    return 1;
  }

  return 0;

}

void close_dingoo (void)
{

  munmap(screen_base, sizeof(unsigned short)*WRES_GOO*HRES_GOO);
  close(fb_dev);
  close(key_dev);

  if (con_dev != -1) {
      ioctl(con_dev, KDSETMODE, KD_TEXT);
      if (con_old_vt != -1) ioctl(con_dev, VT_ACTIVATE, con_old_vt);
          close(con_dev);
          con_dev = -1;
  }

}

inline void linux_draw_bit (Bitu16 *p, Bitu8 v)
{
  *p = pituka_pal[v];
}


void CleanScreen (void)
{
  memset(screen_base, 0x00, sizeof(unsigned short)*WRES_GOO*HRES_GOO);
}

void UpdateScreen_alt (void)
{

}

void LockScreen (int status)
{
  switch (status)
  {
     case 1:
      if (SDL_MUSTLOCK(screen)) {
        if ( SDL_LockSurface(screen) < 0 ){ //Bloqueo el Surface
          //Si da 0 o -1 no puede bloquear el surface
          return;
        }
      }
    break;
    case 0:
      if (SDL_MUSTLOCK(screen)) {
        SDL_UnlockSurface(screen);
      }
    break;
  }

}

void input_bind(Bitu8 key, int pos) //estropea el array de button :?
{

  if(pos > DINGOO_JOY_START)
      return;

  button[pos] = key;

}

Bitu8 event2key (struct input_event *event)
{

		switch (event->code) {
			case KEY_UP: return button[DINGOO_JOY_UP];
			case KEY_LEFT: return button[DINGOO_JOY_LEFT];
			case KEY_RIGHT: return button[DINGOO_JOY_RIGHT];
			case KEY_DOWN: return button[DINGOO_JOY_DOWN];
			case KEY_LEFTCTRL: return button[DINGOO_JOY_A];
			case KEY_LEFTALT: return button[DINGOO_JOY_B];
			case KEY_SPACE: return button[DINGOO_JOY_X];
			case KEY_LEFTSHIFT: return button[DINGOO_JOY_Y];
			case KEY_TAB: return button[DINGOO_JOY_RL];
			case KEY_BACKSPACE: return button[DINGOO_JOY_RR];
			case KEY_ENTER: 
				if(EscDone) emuDone = 1;
				return button[DINGOO_JOY_START];
			case KEY_ESC: 
				if(event->value) {
					CPC.control = 1;
					EscDone = 1;
					VDU.vstart = 77;
					VDU.vheight = 175;
				}else
					EscDone = 0;

			default: return 0xff;
		}

}


void poll_input (void) {
     Bitu8 cpc_key;

	//struct input_event ev;
	if (key_dev == -1 || read(key_dev, &KeyEv, sizeof(KeyEv)) != sizeof(KeyEv)) return;

	if (KeyEv.type == EV_KEY) {

          if(CPC.control) //keyboard on screen
          {
              Keyboard_getControl();
              return;
          }

          switch(KeyEv.value)
          {
	    case 1:
                  cpc_key = event2key(&KeyEv); // translate the PC key to a CPC key
                  if (cpc_key != 0xff) {
                     keyboard_matrix[cpc_key >> 4] &= ~bit_values[cpc_key & 7]; // key is being held down
                  }
                break;
            case 0:
                  cpc_key = event2key(&KeyEv); // translate the PC key to a CPC key
                  if (cpc_key != 0xff) {
                        keyboard_matrix[cpc_key >> 4] |= bit_values[cpc_key & 7]; // key has been released
                  }
                break;
          }

	}

   return;
}

int main(int argc, char *argv[]) {
  int i,length;

   if(!init_dingoo())
     return 1;

   if(!init_buffer())
     return 2;

   for (i = 1; i < argc; i++) { // loop for all command line arguments
      length = strlen(argv[i]);
      if (length > 5 && length < 256) { // minumum for a valid filename
         if (argv[i][0] == '"') { // argument passed with quotes?
            length -= 2;
            strncpy(globalRom.filename, &argv[i][1], length); // strip the quotes
         } else {
            strncpy(globalRom.filename, &argv[i][0], sizeof(rom_name)-1); // take it as is
         }
         printf("ROM: %s\n", globalRom.filename);
 	 WiiStatus.LoadDISK = true;
         globalRom.location = SU_SD;
      }
  }

  InitGFX();

  ShowSplash();
  CleanScreen ();

  cpc_main();

  close_buffer();
  close_dingoo();

  SDL_Quit();
  FreeGFX();

  return 0;
}

int load_rom(t_WiiRom * romfs)
{
   int fileSize = 0;
   void * fbuffer = NULL;
   FILE *pfile;

          //strcpy(romfs->filename, rom_name);

	  if(!(pfile = fopen(romfs->filename, "rb"))){
	        spool = 1;
	    	sprintf(spool_cad, " LOAD ROM: ERROR (%s)", romfs->filename);
		return false;
	  }

	  fseek(pfile, 0, SEEK_END);
	  fileSize = ftell(pfile);
	  rewind(pfile);

	  fbuffer = (void *) malloc(fileSize); 

	  if(fbuffer == NULL){
	        spool = 1;
	    	sprintf(spool_cad, " LOAD ROM: NO MEMORY (%i)", fileSize);
	        fclose(pfile);
		return false;
	  }

	  if(!fread(fbuffer, 1, fileSize, pfile)){
		free(fbuffer);
		fclose(pfile);
	        spool = 1;
	    	sprintf(spool_cad, " LOAD ROM: NO READED? (%s)", romfs->filename);
		return false;
	  }

	  fclose(pfile);

  bool result = false;

  printf("FILE - SDREAD: %s\n",romfs->filename);

  if(!(result = loadBuffered_rom (fbuffer, fileSize))){
        spool = 1;
    	sprintf(spool_cad, " LOAD BUFFER ROM: BAD? (%s)", romfs->filename);
  }
 
  free(fbuffer);

  return result;
}

unsigned int GetTicks(void) { return (unsigned int) SDL_GetTicks(); }

int init_buffer (void)
{

  pbGPBuffer = (Bitu8*) malloc(128*1024); //need by unzip lib

  pix = screen_base; //(void *)malloc(4 * WRES * HRES);
  if(pix == NULL)
      return 0;

  return 1;
}

void close_buffer (void)
{
  if(pix != NULL)
    {
      //free(pix);
      pix = NULL;
    }

  free(pbGPBuffer);
}

extern t_img keyboard;

void ShowEmuCommon (void)
{
  if(FDC.led)
    ShowLed();

  if(CPC.control)
    ShowKeyboard();

}

void ShowMenu (int nMenu) 
{ 
/*
 while(1){
  FillScreenPal((unsigned char *)pituka_arranque);  //show splash

  if(linux_input()) {
*/
	emu_paused (0);
/*
        break;
  }
 }
*/
}

void main_process_pause (int val) {}

void UpdateScreen(void){}
void FillScreen( int Updated ){ }

